import pytest
from django.conf import settings

from getpaid import FraudStatus, PaymentStatus
from getpaid.processor import BaseProcessor
from getpaid.registry import registry

from .tools import Plugin

dummy = 'getpaid.backends.dummy'


class TestRegistry:
    @pytest.fixture(autouse=True)
    def setup_plugin(self):
        if Plugin.slug not in registry:
            registry.register(Plugin)

    def test_register(self):
        assert dummy in settings.INSTALLED_APPS
        assert dummy in registry
        assert issubclass(registry[dummy], BaseProcessor)
        assert Plugin.slug in registry

    def test_get_choices(self):
        choices = registry.get_choices('USD')
        assert len(choices) >= 1
        slugs = [c[0] for c in choices]
        assert Plugin.slug in slugs

    def test_url(self):
        assert len(registry.urls) >= 0

    def test_get_all_supported_currency_choices(self):
        choices = registry.get_all_supported_currency_choices()
        currency_codes = {code for code, _label in choices}
        assert 'EUR' in currency_codes
        assert 'USD' in currency_codes
        for code, label in choices:
            assert code == label
            assert len(code) == 3
            assert code == code.upper()

    def test_choices(self):
        fraud_choices = FraudStatus.CHOICES
        assert type(fraud_choices) == tuple

        payment_choices = PaymentStatus.CHOICES
        assert type(payment_choices) == tuple



"""Dummy payment backend for development and testing.

Self-contained processor simulating payment flows without external
HTTP calls. Supports REST/POST/GET initiation and PUSH/PULL confirmation.

Settings (via GETPAID_BACKEND_SETTINGS['getpaid.backends.dummy']):
    paywall_method: 'REST' | 'POST' | 'GET' (default: 'REST')
    confirmation_method: 'PUSH' | 'PULL' (default: 'PUSH')
"""

import json
import logging
from decimal import Decimal

from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseRedirect,
)
from django.template.response import TemplateResponse
from getpaid_core.fsm import create_payment_machine
from transitions.core import MachineError

from getpaid.post_forms import PaymentHiddenInputsPostForm
from getpaid.processor import BaseProcessor
from getpaid.types import PaymentStatus as ps

logger = logging.getLogger(__name__)


def _can_trigger(payment, trigger_name):
    """Check if a trigger can proceed without actually firing it."""
    trigger = getattr(payment, trigger_name, None)
    if trigger is None:
        return False
    try:
        return payment.may_trigger(trigger_name)
    except (AttributeError, MachineError):
        # If machine not attached or method unavailable
        return False


class PaymentProcessor(BaseProcessor):
    slug = 'dummy'
    display_name = 'Dummy'
    accepted_currencies = ['PLN', 'EUR']
    ok_statuses = [200]
    method = 'REST'
    confirmation_method = 'PUSH'
    post_form_class = PaymentHiddenInputsPostForm
    post_template_name = 'getpaid_dummy/payment_post_form.html'

    def get_paywall_method(self):
        return self.get_setting('paywall_method', self.method)

    def get_confirmation_method(self):
        return self.get_setting(
            'confirmation_method', self.confirmation_method
        ).upper()

    def prepare_transaction(self, request=None, view=None, **kwargs):
        """Simulate payment preparation. No external HTTP calls."""
        method = self.get_paywall_method()

        # Ensure FSM is attached
        create_payment_machine(self.payment)
        self.payment.confirm_prepared()  # ty: ignore[possibly-missing-attribute]
        self.payment.save()  # ty: ignore[possibly-missing-attribute]

        if method == 'POST':
            params = {
                'amount': str(self.payment.amount_required),
                'currency': self.payment.currency,
                'description': self.payment.description,
            }
            form = self.get_form(params)
            return TemplateResponse(
                request=request,
                template=self.get_template_names(view=view),
                context={'form': form, 'paywall_url': '#dummy'},
            )

        redirect_url = self.get_our_baseurl(request)
        return HttpResponseRedirect(redirect_url)

    def handle_paywall_callback(self, request, **kwargs):
        """Handle a simulated callback with JSON body."""
        try:
            data = json.loads(request.body)
        except (json.JSONDecodeError, ValueError):
            return HttpResponseBadRequest('Invalid JSON')

        new_status = data.get('new_status')
        if new_status is None:
            return HttpResponseBadRequest('Missing new_status')

        create_payment_machine(self.payment)

        if new_status == ps.FAILED:
            self.payment.fail()  # ty: ignore[possibly-missing-attribute]
        elif new_status == ps.PRE_AUTH:
            self.payment.confirm_lock()  # ty: ignore[possibly-missing-attribute]
        elif new_status == ps.PAID:
            if _can_trigger(self.payment, 'confirm_lock'):
                self.payment.confirm_lock()  # ty: ignore[possibly-missing-attribute]
            if _can_trigger(self.payment, 'confirm_payment'):
                self.payment.confirm_payment()  # ty: ignore[possibly-missing-attribute]
            if _can_trigger(self.payment, 'mark_as_paid'):
                try:
                    self.payment.mark_as_paid()  # ty: ignore[possibly-missing-attribute]
                except MachineError:
                    logger.debug(
                        'Cannot mark as paid (guard failed).',
                        extra={
                            'payment_id': self.payment.id,
                        },
                    )
        else:
            return HttpResponseBadRequest(f'Unhandled status: {new_status}')

        self.payment.save()  # ty: ignore[possibly-missing-attribute]
        return HttpResponse('OK')

    def fetch_payment_status(self, **kwargs):
        """Simulate fetching status from external provider."""
        status = self.payment.status
        simulated = self.get_setting('confirmation_status', 'paid')

        if status in (
            ps.PAID,
            ps.FAILED,
            ps.REFUNDED,
            ps.REFUND_STARTED,
        ):
            return {}

        if status == ps.NEW:
            return {}

        if simulated == 'failed':
            return {'callback': 'fail'}
        if simulated == 'pre_auth':
            return {'callback': 'confirm_lock'}
        return {
            'callback': 'confirm_payment',
            'amount': self.payment.amount_required,
        }

    def charge(self, amount=None, **kwargs):
        """Simulate charging a pre-authorized amount."""
        if amount is None:
            amount = self.payment.amount_locked
        return {
            'amount_charged': Decimal(str(amount)),
            'success': True,
        }

    def release_lock(self, **kwargs):
        """Simulate releasing a locked amount."""
        return Decimal(str(self.payment.amount_locked))

    def start_refund(self, amount=None, **kwargs):
        """Simulate starting a refund."""
        if amount is None:
            amount = self.payment.amount_paid
        return Decimal(str(amount))

    def cancel_refund(self, **kwargs):
        """Simulate cancelling a refund."""
        return True



"""Abstract models for django-getpaid v3.

Models are plain Django models. FSM transitions are NOT defined on the model.
Instead, use getpaid_core.fsm.create_payment_machine(payment) to attach
transition triggers at runtime.
"""

import logging
import uuid
from decimal import Decimal
from importlib import import_module

import swapper
from django import forms
from django.db import models
from django.db.transaction import atomic
from django.forms import BaseForm
from django.http import HttpRequest, HttpResponse
from django.shortcuts import resolve_url
from django.utils.translation import gettext_lazy as _
from django.views import View
from getpaid_core.fsm import (
    ALLOWED_CALLBACKS,
    create_payment_machine,
)

from getpaid.exceptions import ChargeFailure
from getpaid.types import (
    BuyerInfo,
    ChargeResponse,
    ItemInfo,
    PaymentStatusResponse,
    RestfulResult,
)
from getpaid.types import FraudStatus as fs
from getpaid.types import PaymentStatus as ps

logger = logging.getLogger(__name__)


class AbstractOrder(models.Model):
    """Base class for Order models.

    Consider using UUIDField as primary or secondary key to hide volume.
    """

    class Meta:
        abstract = True

    def get_return_url(
        self,
        *args,
        success: bool | None = None,
        **kwargs,
    ) -> str:
        """Return URL after payment completion.

        Override to customize. Default: get_absolute_url().
        """
        return self.get_absolute_url()

    def get_absolute_url(self) -> str:
        raise NotImplementedError

    def is_ready_for_payment(self) -> bool:
        """Validate order is ready for payment.

        Override for custom validation. Raise ValidationError
        for verbose error messages.
        """
        if self.payments.exclude(status=ps.FAILED).exists():  # ty: ignore[unresolved-attribute]
            raise forms.ValidationError(
                _('Non-failed Payments exist for this Order.')
            )
        return True

    def get_items(self) -> list[ItemInfo]:
        """Return list of items for the payment."""
        return [
            {
                'name': self.get_description(),
                'quantity': 1,
                'unit_price': self.get_total_amount(),
            }
        ]

    def get_total_amount(self) -> Decimal:
        raise NotImplementedError

    def get_buyer_info(self) -> BuyerInfo:
        raise NotImplementedError

    def get_description(self) -> str:
        raise NotImplementedError


class AbstractPayment(models.Model):
    """Abstract payment model.

    FSM transitions are NOT defined on the model. Instead, use
    getpaid_core.fsm.create_payment_machine(payment) to attach
    transition triggers at runtime.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    order = models.ForeignKey(
        swapper.get_model_name('getpaid', 'Order'),
        verbose_name=_('order'),
        on_delete=models.CASCADE,
        related_name='payments',
    )
    amount_required = models.DecimalField(
        _('amount required'),
        decimal_places=2,
        max_digits=20,
        help_text=_(
            'Amount required to fulfill the payment; '
            'in selected currency, normal notation'
        ),
    )
    currency = models.CharField(_('currency'), max_length=3)
    status = models.CharField(
        _('status'),
        max_length=50,
        choices=ps.choices,
        default=ps.NEW,
        db_index=True,
    )
    backend = models.CharField(_('backend'), max_length=100, db_index=True)
    created_on = models.DateTimeField(
        _('created on'), auto_now_add=True, db_index=True
    )
    last_payment_on = models.DateTimeField(
        _('paid on'),
        blank=True,
        null=True,
        default=None,
        db_index=True,
    )
    amount_locked = models.DecimalField(
        _('amount locked'),
        decimal_places=2,
        max_digits=20,
        default=0,
        help_text=_('Amount locked with this payment, ready to charge.'),
    )
    amount_paid = models.DecimalField(
        _('amount paid'),
        decimal_places=2,
        max_digits=20,
        default=0,
        help_text=_('Amount actually paid.'),
    )
    refunded_on = models.DateTimeField(
        _('refunded on'),
        blank=True,
        null=True,
        default=None,
        db_index=True,
    )
    amount_refunded = models.DecimalField(
        _('amount refunded'),
        decimal_places=2,
        max_digits=20,
        default=0,
    )
    external_id = models.CharField(
        _('external id'),
        max_length=64,
        blank=True,
        db_index=True,
        default='',
    )
    description = models.CharField(
        _('description'), max_length=128, blank=True, default=''
    )
    fraud_status = models.CharField(
        _('fraud status'),
        max_length=20,
        choices=fs.choices,
        default=fs.UNKNOWN,
        db_index=True,
    )
    fraud_message = models.TextField(_('fraud message'), blank=True)

    class Meta:
        abstract = True
        ordering = ['-created_on']
        verbose_name = _('Payment')
        verbose_name_plural = _('Payments')

    def __str__(self):
        return f'Payment #{self.id}'

    # ---- Properties ----

    @property
    def fully_paid(self) -> bool:
        return self.amount_paid >= self.amount_required

    def is_fully_paid(self) -> bool:
        """Check if payment is fully paid. Used by core FSM guard."""
        return self.amount_paid >= self.amount_required

    def is_fully_refunded(self) -> bool:
        """Check if payment is fully refunded. Used by core FSM guard."""
        return (
            self.amount_refunded > 0
            and self.amount_refunded >= self.amount_paid
        )

    # ---- Delegation helpers ----

    def get_unique_id(self) -> str:
        """Return unique identifier for this payment."""
        return str(self.id)

    def get_items(self) -> list[ItemInfo]:
        """Relay to order's get_items()."""
        return self.order.get_items()  # ty: ignore[possibly-missing-attribute]

    def get_buyer_info(self) -> BuyerInfo:
        return self.order.get_buyer_info()  # ty: ignore[possibly-missing-attribute]

    def _get_processor(self):
        """Get processor instance for this payment's backend."""
        # Circular import workaround: registry imports processor
        from getpaid.registry import registry

        if self.backend in registry:
            processor_class = registry[self.backend]
        else:
            # last resort if backend removed from INSTALLED_APPS
            module = import_module(str(self.backend))
            processor_class = module.PaymentProcessor
        return processor_class(self)

    def get_form(self, *args, **kwargs) -> BaseForm:
        """Interfaces processor's get_form."""
        return self._get_processor().get_form(*args, **kwargs)

    def get_template_names(self, view: View | None = None) -> list[str]:
        """Interfaces processor's get_template_names."""
        return self._get_processor().get_template_names(view=view)

    def handle_paywall_callback(
        self, request: HttpRequest, **kwargs
    ) -> HttpResponse:
        """Interfaces processor's handle_paywall_callback."""
        return self._get_processor().handle_paywall_callback(request, **kwargs)

    def fetch_status(self) -> PaymentStatusResponse:
        """Interfaces processor's fetch_payment_status."""
        return self._get_processor().fetch_payment_status()

    @atomic
    def fetch_and_update_status(self) -> PaymentStatusResponse:
        """Used during PULL flow to fetch and update payment status."""
        status_report = self.fetch_status()
        callback_name = status_report.get('callback')
        if callback_name:
            if callback_name not in ALLOWED_CALLBACKS:
                logger.warning(
                    'Disallowed callback %r requested for payment %s.',
                    callback_name,
                    self.id,
                    extra={
                        'payment_id': self.id,
                        'payment_status': self.status,
                        'callback': callback_name,
                    },
                )
                status_report['exception'] = ValueError(
                    f'Disallowed callback: {callback_name!r}'
                )
                return status_report

            # Attach FSM and try the callback
            create_payment_machine(self)
            callback = getattr(self, callback_name, None)
            amount = status_report.get('amount', None)
            if callback is not None and self.may_trigger(callback_name):  # ty: ignore[unresolved-attribute]
                try:
                    status_report['callback_result'] = callback(amount=amount)
                    self.save()
                    status_report['saved'] = True
                except Exception as exc:
                    status_report['exception'] = exc
            else:
                logger.debug(
                    'Cannot run fetch+update callback %s.',
                    callback_name,
                    extra={
                        'payment_id': self.id,
                        'payment_status': self.status,
                        'callback': callback_name,
                    },
                )
        return status_report

    def get_return_redirect_url(
        self, request: HttpRequest, success: bool
    ) -> str:
        """Determine redirect URL after payment."""
        processor = self._get_processor()
        if success:
            url = processor.get_setting('SUCCESS_URL')
        else:
            url = processor.get_setting('FAILURE_URL')

        if url is not None:
            kwargs = self.get_return_redirect_kwargs(request, success)
            return resolve_url(url, **kwargs)
        return resolve_url(self.order.get_return_url(self, success=success))  # ty: ignore[possibly-missing-attribute]

    def get_return_redirect_kwargs(
        self, request: HttpRequest, success: bool
    ) -> dict:
        return {'pk': self.id}

    # ---- Action helpers (delegate to processor) ----

    def prepare_transaction(
        self,
        request: HttpRequest | None = None,
        view: View | None = None,
        **kwargs,
    ) -> HttpResponse:
        """Interfaces processor's prepare_transaction."""
        return self._get_processor().prepare_transaction(
            request=request, view=view, **kwargs
        )

    def prepare_transaction_for_rest(
        self,
        request: HttpRequest | None = None,
        view: View | None = None,
        **kwargs,
    ) -> RestfulResult:
        """Helper returning data as dict for REST integration."""
        result = self.prepare_transaction(request=request, view=view, **kwargs)
        data = {'status_code': result.status_code, 'result': result}
        if result.status_code == 200:
            data['target_url'] = result.context_data['paywall_url']  # ty: ignore[unresolved-attribute]
            data['form'] = {
                'fields': [
                    {
                        'name': name,
                        'value': field.initial,
                        'label': field.label or name,
                        'widget': field.widget.__class__.__name__,
                        'help_text': field.help_text,
                        'required': field.required,
                    }
                    for name, field in result.context_data[  # ty: ignore[unresolved-attribute]
                        'form'
                    ].fields.items()
                ],
            }
        elif result.status_code == 302:
            data['target_url'] = result.url  # ty: ignore[unresolved-attribute]
        else:
            data['message'] = result.content
        return data  # ty: ignore[invalid-return-type]

    @atomic
    def charge(
        self,
        amount: Decimal | float | int | None = None,
        **kwargs,
    ) -> ChargeResponse:
        """Interfaces processor's charge with FSM transitions."""
        create_payment_machine(self)
        if amount is None:
            amount = self.amount_locked  # ty: ignore[invalid-assignment]
        if amount > self.amount_locked:
            raise ValueError('Cannot charge more than locked value.')
        processor = self._get_processor()
        result = processor.charge(amount=amount, **kwargs)
        if 'amount_charged' in result or result.get('success', False):
            amount_charged = result.get('amount_charged', amount)
            self.amount_locked -= amount_charged
            self.confirm_payment(amount=amount_charged)  # ty: ignore[unresolved-attribute]
            if self.may_trigger('mark_as_paid'):  # ty: ignore[unresolved-attribute]
                try:
                    self.mark_as_paid()  # ty: ignore[unresolved-attribute]
                except Exception:
                    logger.debug(
                        'Cannot mark as fully paid, left as partially paid.',
                        extra={
                            'payment_id': self.id,
                            'payment_status': self.status,
                        },
                    )
        elif result.get('async_call', False):
            if self.may_trigger('confirm_charge_sent'):  # ty: ignore[unresolved-attribute]
                self.confirm_charge_sent()  # ty: ignore[unresolved-attribute]
            else:
                logger.debug(
                    'Cannot confirm charge sent.',
                    extra={
                        'payment_id': self.id,
                        'payment_status': self.status,
                    },
                )
        else:
            raise ChargeFailure(
                'Error occurred while trying to charge locked amount.'
            )
        self.save()
        return result



"""Tests for AbstractPayment methods (via concrete CustomPayment model)."""

import logging
from collections import OrderedDict
from decimal import Decimal
from unittest.mock import MagicMock, patch

import pytest
import swapper
from django import forms
from getpaid_core.fsm import create_fraud_machine, create_payment_machine

pytestmark = pytest.mark.django_db

Order = swapper.load_model('getpaid', 'Order')
Payment = swapper.load_model('getpaid', 'Payment')


def _make_payment(**kwargs):
    """Helper to create an Order + Payment pair."""
    order = Order.objects.create()
    defaults = {
        'order': order,
        'currency': order.currency,
        'amount_required': Decimal(str(order.get_total_amount())),
        'backend': 'getpaid.backends.dummy',
        'description': order.get_description(),
    }
    defaults.update(kwargs)
    return Payment.objects.create(**defaults)


class TestPrepareTransaction:
    """Tests for prepare_transaction and prepare_transaction_for_rest."""

    def test_prepare_transaction_passes_view_to_processor(self):
        """The view parameter should be forwarded to the processor."""
        payment = _make_payment()
        mock_view = MagicMock()
        mock_request = MagicMock()

        proc = payment._get_processor()
        with (
            patch.object(
                type(proc),
                'prepare_transaction',
                return_value=MagicMock(status_code=302, url='/redirect'),
            ) as mock_prep,
            patch.object(type(payment), '_get_processor', return_value=proc),
        ):
            payment.prepare_transaction(request=mock_request, view=mock_view)
            mock_prep.assert_called_once_with(
                request=mock_request, view=mock_view
            )

    def test_prepare_transaction_for_rest_iterates_form_fields(self):
        """Iterating a Django form's .fields dict yields (name, field) tuples."""
        payment = _make_payment()

        mock_form = MagicMock()
        mock_form.fields = OrderedDict([
            ('amount', forms.DecimalField(initial='100.00', label='Amount')),
            ('currency', forms.CharField(initial='EUR', label='Currency')),
        ])

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.context_data = {
            'paywall_url': 'https://example.com/pay',
            'form': mock_form,
        }

        with patch.object(
            type(payment),
            'prepare_transaction',
            return_value=mock_response,
        ):
            result = payment.prepare_transaction_for_rest()

        assert result['status_code'] == 200
        assert result['target_url'] == 'https://example.com/pay'
        assert len(result['form']['fields']) == 2
        assert result['form']['fields'][0]['name'] == 'amount'
        assert result['form']['fields'][1]['name'] == 'currency'


class TestChargeDoubleCount:
    """Tests for the charge() method and its interaction with confirm_payment()."""

    def test_charge_does_not_double_count_amount_paid(self):
        """charge() should not double-count amount_paid."""
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_lock(amount=Decimal('100.00'))
        payment.save()

        proc = payment._get_processor()
        with (
            patch.object(
                type(proc),
                'charge',
                return_value={
                    'amount_charged': Decimal('100.00'),
                    'success': True,
                },
            ),
            patch.object(type(payment), '_get_processor', return_value=proc),
        ):
            payment.charge(amount=Decimal('100.00'))

        assert payment.amount_paid == Decimal('100.00')

    def test_charge_partial_amount(self):
        """Charging a partial amount should correctly update amount_paid."""
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_lock(amount=Decimal('100.00'))
        payment.save()

        proc = payment._get_processor()
        with (
            patch.object(
                type(proc),
                'charge',
                return_value={
                    'amount_charged': Decimal('50.00'),
                    'success': True,
                },
            ),
            patch.object(type(payment), '_get_processor', return_value=proc),
        ):
            payment.charge(amount=Decimal('50.00'))

        assert payment.amount_paid == Decimal('50.00')
        assert payment.amount_locked == Decimal('50.00')


class TestFetchAndUpdateStatusAllowlist:
    """Tests for the FSM callback allowlist in fetch_and_update_status()."""

    def test_payment_has_allowed_callbacks(self):
        """AbstractPayment should reference ALLOWED_CALLBACKS."""
        from getpaid_core.fsm import ALLOWED_CALLBACKS

        assert isinstance(ALLOWED_CALLBACKS, frozenset)

    def test_allowed_callbacks_contains_known_methods(self):
        """ALLOWED_CALLBACKS should contain all known FSM transition method names."""
        from getpaid_core.fsm import ALLOWED_CALLBACKS

        expected = {
            'confirm_prepared',
            'confirm_lock',
            'confirm_charge_sent',
            'confirm_payment',
            'mark_as_paid',
            'fail',
            'confirm_refund',
            'mark_as_refunded',
        }
        assert expected.issubset(ALLOWED_CALLBACKS)

    def test_fetch_and_update_rejects_disallowed_callback(self):
        """fetch_and_update_status() should reject callback names not in allowlist."""
        payment = _make_payment()

        with patch.object(
            type(payment),
            'fetch_status',
            return_value={'callback': 'delete', 'amount': Decimal('100.00')},
        ):
            result = payment.fetch_and_update_status()

        assert 'callback_result' not in result
        assert 'exception' in result

    def test_fetch_and_update_accepts_allowed_callback(self):
        """fetch_and_update_status() should accept callback names in the allowlist."""
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.save()

        with patch.object(
            type(payment),
            'fetch_status',
            return_value={
                'callback': 'confirm_lock',
                'amount': Decimal('100.00'),
            },
        ):
            result = payment.fetch_and_update_status()

        assert 'callback_result' in result or 'saved' in result

    def test_fetch_and_update_logs_disallowed_callback(self, caplog):
        """fetch_and_update_status() should log a warning for disallowed callbacks."""
        payment = _make_payment()

        with caplog.at_level(logging.WARNING, logger='getpaid.abstracts'):
            with patch.object(
                type(payment),
                'fetch_status',
                return_value={
                    'callback': 'save',
                    'amount': Decimal('100.00'),
                },
            ):
                payment.fetch_and_update_status()

        assert any(
            'Disallowed callback' in record.message for record in caplog.records
        )


class TestFraudMethodNames:
    """Tests that fraud methods are accessible with public names (no name mangling)."""

    def test_payment_has_flag_as_fraud(self):
        payment = _make_payment()
        create_fraud_machine(payment)
        assert hasattr(payment, 'flag_as_fraud')
        assert callable(payment.flag_as_fraud)

    def test_payment_has_flag_as_legit(self):
        payment = _make_payment()
        create_fraud_machine(payment)
        assert hasattr(payment, 'flag_as_legit')
        assert callable(payment.flag_as_legit)

    def test_payment_has_flag_for_check(self):
        payment = _make_payment()
        create_fraud_machine(payment)
        assert hasattr(payment, 'flag_for_check')
        assert callable(payment.flag_for_check)

    def test_flag_as_fraud_transitions_from_unknown(self):
        from getpaid.types import FraudStatus as frs

        payment = _make_payment()
        create_fraud_machine(payment)
        assert payment.fraud_status == frs.UNKNOWN
        payment.flag_as_fraud(message='Suspicious IP')
        payment.save()
        assert payment.fraud_status == frs.REJECTED
        assert payment.fraud_message == 'Suspicious IP'

    def test_flag_as_legit_transitions_from_unknown(self):
        from getpaid.types import FraudStatus as frs

        payment = _make_payment()
        create_fraud_machine(payment)
        assert payment.fraud_status == frs.UNKNOWN
        payment.flag_as_legit(message='Verified buyer')
        payment.save()
        assert payment.fraud_status == frs.ACCEPTED
        assert payment.fraud_message == 'Verified buyer'

    def test_flag_for_check_transitions_from_unknown(self):
        from getpaid.types import FraudStatus as frs

        payment = _make_payment()
        create_fraud_machine(payment)
        assert payment.fraud_status == frs.UNKNOWN
        payment.flag_for_check(message='Needs review')
        payment.save()
        assert payment.fraud_status == frs.CHECK
        assert payment.fraud_message == 'Needs review'

    def test_no_triple_underscore_methods(self):
        payment = _make_payment()
        assert not hasattr(payment, '___mark_as_fraud')
        assert not hasattr(payment, '___mark_as_legit')
        assert not hasattr(payment, '___mark_for_check')


class TestAmountRefundedPrecision:
    """Tests that amount_refunded has consistent decimal precision with other money fields."""

    def test_amount_refunded_has_2_decimal_places(self):
        field = Payment._meta.get_field('amount_refunded')
        assert field.decimal_places == 2

    def test_all_money_fields_same_precision(self):
        money_fields = [
            'amount_required',
            'amount_locked',
            'amount_paid',
            'amount_refunded',
        ]
        precisions = {}
        for name in money_fields:
            field = Payment._meta.get_field(name)
            precisions[name] = field.decimal_places
        unique_precisions = set(precisions.values())
        assert len(unique_precisions) == 1, (
            f'Inconsistent decimal places across money fields: {precisions}'
        )


class TestFieldVerboseNames:
    def test_amount_locked_verbose_name(self):
        field = Payment._meta.get_field('amount_locked')
        verbose = str(field.verbose_name).lower()
        assert 'locked' in verbose

    def test_amount_paid_verbose_name(self):
        field = Payment._meta.get_field('amount_paid')
        verbose = str(field.verbose_name).lower()
        assert 'paid' in verbose

    def test_no_duplicate_verbose_names_on_money_fields(self):
        money_fields = [
            'amount_required',
            'amount_locked',
            'amount_paid',
            'amount_refunded',
        ]
        names = {}
        for name in money_fields:
            field = Payment._meta.get_field(name)
            verbose = str(field.verbose_name)
            names[name] = verbose
        unique_names = set(names.values())
        assert len(unique_names) == len(money_fields)


class TestValidatorOrdering:
    def test_validators_run_in_deterministic_order(self, settings):
        from getpaid.validators import run_getpaid_validators

        call_order = []

        def make_validator(name):
            def validator(data):
                call_order.append(name)
                return data

            return validator

        import sys
        import types

        fake_module = types.ModuleType('fake_validators')
        fake_module.v1 = make_validator('v1')
        fake_module.v2 = make_validator('v2')
        fake_module.v3 = make_validator('v3')

        sys.modules['fake_validators'] = fake_module

        try:
            settings.GETPAID = {
                'VALIDATORS': [
                    'fake_validators.v1',
                    'fake_validators.v2',
                ],
                'BACKENDS': {
                    'test_backend': {
                        'VALIDATORS': [
                            'fake_validators.v3',
                        ],
                    },
                },
            }

            data = {'backend': 'test_backend'}
            for _ in range(5):
                call_order.clear()
                run_getpaid_validators(data)
                if _ == 0:
                    first_order = call_order.copy()
                else:
                    assert call_order == first_order

        finally:
            del sys.modules['fake_validators']

    def test_validators_no_duplicates(self, settings):
        from getpaid.validators import run_getpaid_validators

        call_count = {'v1': 0}

        def counting_validator(data):
            call_count['v1'] += 1
            return data

        import sys
        import types

        fake_module = types.ModuleType('fake_validators2')
        fake_module.v1 = counting_validator
        sys.modules['fake_validators2'] = fake_module

        try:
            settings.GETPAID = {
                'VALIDATORS': ['fake_validators2.v1'],
                'BACKENDS': {
                    'test_backend': {
                        'VALIDATORS': ['fake_validators2.v1'],
                    },
                },
            }
            data = {'backend': 'test_backend'}
            run_getpaid_validators(data)
            assert call_count['v1'] == 1
        finally:
            del sys.modules['fake_validators2']



"""Tests for the dummy payment backend (rewritten from scratch)."""

import pathlib
from decimal import Decimal

import pytest
import swapper
from getpaid_core.fsm import create_payment_machine

from getpaid.types import PaymentStatus as ps

pytestmark = pytest.mark.django_db

Order = swapper.load_model('getpaid', 'Order')
Payment = swapper.load_model('getpaid', 'Payment')


def _make_payment(**kwargs):
    """Helper to create an Order + Payment pair."""
    order = Order.objects.create()
    defaults = {
        'order': order,
        'currency': order.currency,
        'amount_required': Decimal(str(order.get_total_amount())),
        'backend': 'getpaid.backends.dummy',
        'description': order.get_description(),
    }
    defaults.update(kwargs)
    return Payment.objects.create(**defaults)


class TestDummyProcessorAttributes:
    def test_slug(self):
        from getpaid.backends.dummy.processor import PaymentProcessor

        assert PaymentProcessor.slug == 'dummy'

    def test_display_name(self):
        from getpaid.backends.dummy.processor import PaymentProcessor

        assert PaymentProcessor.display_name == 'Dummy'

    def test_accepted_currencies(self):
        from getpaid.backends.dummy.processor import PaymentProcessor

        assert 'PLN' in PaymentProcessor.accepted_currencies
        assert 'EUR' in PaymentProcessor.accepted_currencies

    def test_no_requests_import(self):
        import importlib

        mod = importlib.import_module('getpaid.backends.dummy.processor')
        source = pathlib.Path(mod.__file__).open().read()
        assert 'import requests' not in source


class TestDummyPrepareTransaction:
    def test_rest_method_returns_redirect(self, rf, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'paywall_method': 'REST'}
        }
        payment = _make_payment()
        request = rf.get('/')
        result = payment.prepare_transaction(request=request)
        assert result.status_code == 302
        assert payment.status == ps.PREPARED

    def test_get_method_returns_redirect(self, rf, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'paywall_method': 'GET'}
        }
        payment = _make_payment()
        request = rf.get('/')
        result = payment.prepare_transaction(request=request)
        assert result.status_code == 302
        assert payment.status == ps.PREPARED

    def test_post_method_returns_template_response(self, rf, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'paywall_method': 'POST'}
        }
        payment = _make_payment()
        request = rf.get('/')
        result = payment.prepare_transaction(request=request)
        assert result.status_code == 200
        assert payment.status == ps.PREPARED

    def test_prepare_works_without_request(self, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'paywall_method': 'REST'}
        }
        payment = _make_payment()
        result = payment.prepare_transaction(request=None)
        assert result.status_code == 302
        assert payment.status == ps.PREPARED


class TestDummyHandleCallback:
    def test_callback_paid(self, rf):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.save()

        request = rf.post(
            '', content_type='application/json', data={'new_status': ps.PAID}
        )
        response = payment.handle_paywall_callback(request)
        assert response.status_code == 200
        assert payment.status == ps.PAID

    def test_callback_pre_auth(self, rf):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.save()

        request = rf.post(
            '',
            content_type='application/json',
            data={'new_status': ps.PRE_AUTH},
        )
        response = payment.handle_paywall_callback(request)
        assert response.status_code == 200
        assert payment.status == ps.PRE_AUTH

    def test_callback_failed(self, rf):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.save()

        request = rf.post(
            '',
            content_type='application/json',
            data={'new_status': ps.FAILED},
        )
        response = payment.handle_paywall_callback(request)
        assert response.status_code == 200
        assert payment.status == ps.FAILED

    def test_callback_no_status_returns_400(self, rf):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.save()

        request = rf.post('', content_type='application/json', data={})
        response = payment.handle_paywall_callback(request)
        assert response.status_code == 400

    def test_callback_unknown_status_returns_400(self, rf):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.save()

        request = rf.post(
            '',
            content_type='application/json',
            data={'new_status': 'nonsense'},
        )
        response = payment.handle_paywall_callback(request)
        assert response.status_code == 400


class TestDummyFetchPaymentStatus:
    def test_new_payment_no_callback(self):
        payment = _make_payment()
        proc = payment._get_processor()
        result = proc.fetch_payment_status()
        assert result.get('callback') is None

    def test_prepared_payment_returns_confirm_payment(self):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.save()
        proc = payment._get_processor()
        result = proc.fetch_payment_status()
        assert result.get('callback') == 'confirm_payment'
        assert result.get('amount') == payment.amount_required

    def test_terminal_paid_returns_no_callback(self):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        full = payment.amount_required
        payment.confirm_payment(amount=full)
        payment.amount_paid = full
        payment.mark_as_paid()
        payment.save()
        proc = payment._get_processor()
        result = proc.fetch_payment_status()
        assert result == {}

    def test_terminal_failed_returns_no_callback(self):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.fail()
        payment.save()
        proc = payment._get_processor()
        result = proc.fetch_payment_status()
        assert result == {}

    def test_confirmation_status_failed(self, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'confirmation_status': 'failed'}
        }
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.save()
        proc = payment._get_processor()
        result = proc.fetch_payment_status()
        assert result.get('callback') == 'fail'

    def test_confirmation_status_pre_auth(self, settings):
        settings.GETPAID_BACKEND_SETTINGS = {
            'getpaid.backends.dummy': {'confirmation_status': 'pre_auth'}
        }
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_prepared()
        payment.save()
        proc = payment._get_processor()
        result = proc.fetch_payment_status()
        assert result.get('callback') == 'confirm_lock'


class TestDummyCharge:
    def test_charge_returns_dict_with_amount(self):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_lock(amount=Decimal('100.00'))
        payment.save()

        proc = payment._get_processor()
        result = proc.charge(amount=Decimal('50.00'))
        assert isinstance(result, dict)
        assert result['amount_charged'] == Decimal('50.00')
        assert result['success'] is True

    def test_charge_uses_locked_amount_when_none(self):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_lock(amount=Decimal('100.00'))
        payment.save()

        proc = payment._get_processor()
        result = proc.charge(amount=None)
        assert result['amount_charged'] == Decimal('100.00')

    def test_charge_returns_dict_not_none(self):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_lock(amount=Decimal('100.00'))
        payment.save()

        proc = payment._get_processor()
        result = proc.charge(amount=Decimal('100.00'))
        assert result is not None


class TestDummyReleaseLock:
    def test_release_lock_returns_decimal(self):
        payment = _make_payment()
        create_payment_machine(payment)
        payment.confirm_lock(amount=Decimal('100.00'))
        payment.save()

        proc = payment._get_processor()
        result = proc.release_lock()
        assert isinstance(result, Decimal)
        assert result == Decimal('100.00')


class TestDummyStartRefund:
    def test_start_refund_returns_amount(self):
        payment = _make_payment()
        create_payment_machine(payment)
        full = payment.amount_required
        payment.confirm_lock(amount=full)
        payment.confirm_payment(amount=full)
        payment.amount_paid = full
        payment.mark_as_paid()
        payment.save()

        proc = payment._get_processor()
        result = proc.start_refund(amount=Decimal('50.00'))
        assert isinstance(result, Decimal)
        assert result == Decimal('50.00')

    def test_start_refund_defaults_to_paid_amount(self):
        payment = _make_payment()
        create_payment_machine(payment)
        full = payment.amount_required
        payment.confirm_lock(amount=full)
        payment.confirm_payment(amount=full)
        payment.amount_paid = full
        payment.mark_as_paid()
        payment.save()

        proc = payment._get_processor()
        result = proc.start_refund()
        assert result == full


class TestDummyCancelRefund:
    def test_cancel_refund_returns_true(self):
        payment = _make_payment()
        proc = payment._get_processor()
        result = proc.cancel_refund()
        assert result is True



"""FSM transition tests using getpaid-core's transitions library."""

import pytest
import swapper
from getpaid_core.fsm import create_payment_machine

from getpaid.status import PaymentStatus as ps

dummy = 'getpaid.backends.dummy'

Order = swapper.load_model('getpaid', 'Order')
Payment = swapper.load_model('getpaid', 'Payment')

pytestmark = pytest.mark.django_db


def _can_trigger(payment, name):
    """Check if trigger can fire without actually firing."""
    try:
        return payment.may_trigger(name)
    except AttributeError:
        return False


def test_fsm_direct_prepare(payment_factory):
    payment = payment_factory()
    create_payment_machine(payment)
    assert payment.status == ps.NEW
    payment.confirm_prepared()
    assert payment.status == ps.PREPARED


def test_fsm_check_available_transitions_from_new(payment_factory):
    payment = payment_factory()
    create_payment_machine(payment)
    assert payment.status == ps.NEW

    assert _can_trigger(payment, 'confirm_prepared')
    assert _can_trigger(payment, 'confirm_lock')
    assert not _can_trigger(payment, 'confirm_charge_sent')
    assert not _can_trigger(payment, 'confirm_payment')
    assert not _can_trigger(payment, 'mark_as_paid')
    assert not _can_trigger(payment, 'release_lock')
    assert not _can_trigger(payment, 'start_refund')
    assert not _can_trigger(payment, 'cancel_refund')
    assert not _can_trigger(payment, 'confirm_refund')
    assert not _can_trigger(payment, 'mark_as_refunded')
    assert _can_trigger(payment, 'fail')


def test_fsm_check_available_transitions_from_failed(payment_factory):
    payment = payment_factory()
    create_payment_machine(payment)
    payment.fail()
    assert payment.status == ps.FAILED

    assert not _can_trigger(payment, 'confirm_prepared')
    assert not _can_trigger(payment, 'confirm_lock')
    assert not _can_trigger(payment, 'confirm_charge_sent')
    assert not _can_trigger(payment, 'confirm_payment')
    assert not _can_trigger(payment, 'mark_as_paid')
    assert not _can_trigger(payment, 'release_lock')
    assert not _can_trigger(payment, 'start_refund')
    assert not _can_trigger(payment, 'cancel_refund')
    assert not _can_trigger(payment, 'confirm_refund')
    assert not _can_trigger(payment, 'mark_as_refunded')
    assert not _can_trigger(payment, 'fail')


def test_fsm_check_available_transitions_from_prepared(payment_factory):
    payment = payment_factory()
    create_payment_machine(payment)
    payment.confirm_prepared()
    assert payment.status == ps.PREPARED

    assert not _can_trigger(payment, 'confirm_prepared')
    assert _can_trigger(payment, 'confirm_lock')
    assert not _can_trigger(payment, 'confirm_charge_sent')
    assert _can_trigger(payment, 'confirm_payment')
    assert not _can_trigger(payment, 'mark_as_paid')
    assert not _can_trigger(payment, 'release_lock')
    assert not _can_trigger(payment, 'start_refund')
    assert not _can_trigger(payment, 'cancel_refund')
    assert not _can_trigger(payment, 'confirm_refund')
    assert not _can_trigger(payment, 'mark_as_refunded')
    assert _can_trigger(payment, 'fail')


def test_fsm_check_available_transitions_from_locked(payment_factory):
    payment = payment_factory()
    create_payment_machine(payment)
    payment.confirm_lock()
    assert payment.status == ps.PRE_AUTH

    assert not _can_trigger(payment, 'confirm_prepared')
    assert not _can_trigger(payment, 'confirm_lock')
    assert _can_trigger(payment, 'confirm_charge_sent')
    assert _can_trigger(payment, 'confirm_payment')
    assert not _can_trigger(payment, 'mark_as_paid')
    assert _can_trigger(payment, 'release_lock')
    assert not _can_trigger(payment, 'start_refund')
    assert not _can_trigger(payment, 'cancel_refund')
    assert not _can_trigger(payment, 'confirm_refund')
    assert not _can_trigger(payment, 'mark_as_refunded')
    assert _can_trigger(payment, 'fail')


def test_fsm_check_available_transitions_from_partial(payment_factory):
    payment = payment_factory()
    create_payment_machine(payment)
    payment.confirm_lock()
    payment.confirm_payment()
    assert payment.status == ps.PARTIAL

    assert not _can_trigger(payment, 'confirm_prepared')
    assert not _can_trigger(payment, 'confirm_lock')
    assert not _can_trigger(payment, 'confirm_charge_sent')
    assert _can_trigger(payment, 'confirm_payment')
    # mark_as_paid has a guard -- may_trigger may return True
    # but actual trigger will raise MachineError if not fully paid
    assert _can_trigger(payment, 'mark_as_paid')
    assert not _can_trigger(payment, 'release_lock')
    assert _can_trigger(payment, 'start_refund')
    assert not _can_trigger(payment, 'cancel_refund')
    assert not _can_trigger(payment, 'confirm_refund')
    assert not _can_trigger(payment, 'fail')
    assert _can_trigger(payment, 'mark_as_refunded')


def test_fsm_check_available_transitions_from_paid(payment_factory):
    payment = payment_factory()
    create_payment_machine(payment)
    payment.confirm_lock()
    payment.confirm_payment()
    # Need to set amount_paid for guard
    payment.amount_paid = payment.amount_required
    payment.mark_as_paid()
    assert payment.status == ps.PAID

    assert not _can_trigger(payment, 'confirm_prepared')
    assert not _can_trigger(payment, 'confirm_lock')
    assert not _can_trigger(payment, 'confirm_charge_sent')
    assert not _can_trigger(payment, 'confirm_payment')
    assert not _can_trigger(payment, 'mark_as_paid')
    assert not _can_trigger(payment, 'release_lock')
    assert _can_trigger(payment, 'start_refund')
    assert not _can_trigger(payment, 'cancel_refund')
    assert not _can_trigger(payment, 'confirm_refund')
    assert not _can_trigger(payment, 'mark_as_refunded')
    assert not _can_trigger(payment, 'fail')



import uuid

import pytest
import swapper
from django.template.response import TemplateResponse
from getpaid_core.fsm import create_payment_machine

from getpaid.registry import registry
from getpaid.types import BackendMethod as bm
from getpaid.types import ConfirmationMethod as cm
from getpaid.types import PaymentStatus as ps

from .tools import Plugin

pytestmark = pytest.mark.django_db
dummy = 'getpaid.backends.dummy'

Order = swapper.load_model('getpaid', 'Order')
Payment = swapper.load_model('getpaid', 'Payment')


def _prep_conf(api_method: bm = bm.REST, confirm_method: cm = cm.PUSH) -> dict:
    return {
        'getpaid.backends.dummy': {
            'paywall_method': api_method,
            'confirmation_method': confirm_method,
        }
    }


class TestModelProcessor:
    @pytest.fixture(autouse=True)
    def setup_plugin(self):
        if Plugin.slug not in registry:
            registry.register(Plugin)

    def test_model_and_dummy_backend(self):
        order = Order.objects.create()
        payment = Payment.objects.create(
            order=order,
            currency=order.currency,
            amount_required=order.get_total_amount(),
            backend=dummy,
            description=order.get_description(),
        )
        proc = payment._get_processor()
        assert isinstance(proc, registry[dummy])

    def test_model_and_test_backend(self):
        order = Order.objects.create()
        payment = Payment.objects.create(
            order=order,
            currency=order.currency,
            amount_required=order.get_total_amount(),
            backend=Plugin.slug,
            description=order.get_description(),
        )
        proc = payment._get_processor()
        assert isinstance(proc, registry[Plugin.slug])


def test_get_flow_begin(payment_factory, settings, rf):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(api_method=bm.GET)
    payment = payment_factory(external_id=uuid.uuid4())

    result = payment.prepare_transaction(None)
    assert result.status_code == 302


def test_post_flow_begin(payment_factory, settings, rf):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(api_method=bm.POST)
    payment = payment_factory(external_id=uuid.uuid4())

    result = payment.prepare_transaction(None)
    assert result.status_code == 200
    assert isinstance(result, TemplateResponse)
    assert payment.status == ps.PREPARED


def test_rest_flow_begin(payment_factory, settings, rf):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(api_method=bm.REST)
    payment = payment_factory(external_id=uuid.uuid4())

    result = payment.prepare_transaction(None)
    assert result.status_code == 302
    assert payment.status == ps.PREPARED


def test_pull_flow_paid(payment_factory, settings):
    settings.GETPAID_BACKEND_SETTINGS = {
        'getpaid.backends.dummy': {
            'confirmation_method': cm.PULL,
            'confirmation_status': 'paid',
        }
    }

    payment = payment_factory(external_id=uuid.uuid4())
    create_payment_machine(payment)
    payment.confirm_prepared()

    payment.fetch_and_update_status()
    assert payment.status == ps.PARTIAL
    create_payment_machine(payment)
    assert payment.may_trigger('mark_as_paid')


def test_pull_flow_locked(payment_factory, settings):
    settings.GETPAID_BACKEND_SETTINGS = {
        'getpaid.backends.dummy': {
            'confirmation_method': cm.PULL,
            'confirmation_status': 'pre_auth',
        }
    }

    payment = payment_factory(external_id=uuid.uuid4())
    create_payment_machine(payment)
    payment.confirm_prepared()

    payment.fetch_and_update_status()
    assert payment.status == ps.PRE_AUTH


def test_pull_flow_failed(payment_factory, settings):
    settings.GETPAID_BACKEND_SETTINGS = {
        'getpaid.backends.dummy': {
            'confirmation_method': cm.PULL,
            'confirmation_status': 'failed',
        }
    }

    payment = payment_factory(external_id=uuid.uuid4())
    create_payment_machine(payment)
    payment.confirm_prepared()

    payment.fetch_and_update_status()
    assert payment.status == ps.FAILED


def test_push_flow_paid(payment_factory, settings, rf):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(confirm_method=cm.PUSH)

    payment = payment_factory(external_id=uuid.uuid4())
    create_payment_machine(payment)
    payment.confirm_prepared()

    request = rf.post(
        '', content_type='application/json', data={'new_status': ps.PAID}
    )
    payment.handle_paywall_callback(request)
    assert payment.status == ps.PAID


def test_push_flow_locked(payment_factory, settings, rf):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(confirm_method=cm.PUSH)

    payment = payment_factory(external_id=uuid.uuid4())
    create_payment_machine(payment)
    payment.confirm_prepared()

    request = rf.post(
        '', content_type='application/json', data={'new_status': ps.PRE_AUTH}
    )
    payment.handle_paywall_callback(request)
    assert payment.status == ps.PRE_AUTH


def test_push_flow_failed(payment_factory, settings, rf):
    settings.GETPAID_BACKEND_SETTINGS = _prep_conf(confirm_method=cm.PUSH)

    payment = payment_factory(external_id=uuid.uuid4())
    create_payment_machine(payment)
    payment.confirm_prepared()

    request = rf.post(
        '', content_type='application/json', data={'new_status': ps.FAILED}
    )
    payment.handle_paywall_callback(request)
    assert payment.status == ps.FAILED
